"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
const version_1 = require("./version");
const base_exports = __importStar(require("./basewidget"));
const uioutput_exports = __importStar(require("./uioutput"));
const uibuilder_exports = __importStar(require("./uibuilder"));
const toolbox_exports = __importStar(require("./toolbox"));
const mainmenu_1 = require("@jupyterlab/mainmenu");
const toolbox_1 = require("./toolbox");
const registry_1 = require("./registry");
const application_1 = require("@jupyterlab/application");
const notebook_1 = require("@jupyterlab/notebook");
const context_1 = require("./context");
const documentation = 'nbtools:documentation';
const all_exports = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, base_exports), uioutput_exports), uibuilder_exports), toolbox_exports), { documentation });
const EXTENSION_ID = '@genepattern/nbtools:plugin';
const NAMESPACE = 'nbtools';
/**
 * The nbtools plugin.
 */
const nbtools_plugin = {
    id: EXTENSION_ID,
    // provides: IToolRegistry,
    requires: [base_1.IJupyterWidgetRegistry],
    optional: [mainmenu_1.IMainMenu, application_1.ILayoutRestorer, application_1.ILabShell, notebook_1.INotebookTracker],
    activate: activate_widget_extension,
    autoStart: true
};
exports.default = nbtools_plugin;
/**
 * Activate the widget extension.
 */
function activate_widget_extension(app, widget_registry, mainmenu, restorer, shell, notebook_tracker) {
    // Create the tool registry
    const tool_registry = new registry_1.ToolRegistry(notebook_tracker);
    // Initialize the ContextManager
    init_context(app, notebook_tracker, tool_registry);
    // Add items to the help menu
    add_documentation_link(app, mainmenu);
    // Add keyboard shortcuts
    add_keyboard_shortcuts(app, tool_registry);
    // Add the toolbox
    add_tool_browser(app, restorer);
    // Register the nbtools widgets with the widget registry
    widget_registry.registerWidget({
        name: version_1.MODULE_NAME,
        version: version_1.MODULE_VERSION,
        exports: all_exports,
    });
    // Return the tool registry so that it is provided to other extensions
    return tool_registry;
}
function init_context(app, notebook_tracker, tool_registry) {
    context_1.ContextManager.jupyter_app = app;
    context_1.ContextManager.notebook_tracker = notebook_tracker;
    context_1.ContextManager.tool_registry = tool_registry;
    //(window as any).ContextManager = ContextManager;  // Left in for development purposes
}
function add_keyboard_shortcuts(app, tool_registry) {
    app.commands.addCommand("nbtools:insert-tool", {
        label: 'Insert Notebook Tool',
        execute: () => {
            // Open the tool manager, if necessary
            app.shell.activateById('nbtools-browser');
            // If only one tool is available, add it
            const tools = tool_registry.list();
            if (tools.length === 1)
                toolbox_1.Toolbox.add_tool_cell(tools[0]);
            // Otherwise give the search box focus
            else
                document.querySelector('.nbtools-search').focus();
        },
    });
}
function add_tool_browser(app, restorer) {
    const tool_browser = new toolbox_1.ToolBrowser();
    tool_browser.title.iconClass = 'nbtools-icon fa fa-th jp-SideBar-tabIcon';
    tool_browser.title.caption = 'Toolbox';
    tool_browser.id = 'nbtools-browser';
    // Add the tool browser widget to the application restorer
    if (restorer)
        restorer.add(tool_browser, NAMESPACE);
    app.shell.add(tool_browser, 'left', { rank: 102 });
}
/**
 * Add the nbtools documentation link to the help menu
 *
 * @param {Application<Widget>} app
 * @param {IMainMenu} mainmenu
 */
function add_documentation_link(app, mainmenu) {
    // Add documentation command to the command palette
    app.commands.addCommand(documentation, {
        label: 'nbtools Documentation',
        caption: 'Open documentation for nbtools',
        isEnabled: () => !!app.shell,
        execute: () => {
            const url = 'https://github.com/genepattern/nbtools#nbtools';
            let element = document.createElement('a');
            element.href = url;
            element.target = '_blank';
            document.body.appendChild(element);
            element.click();
            document.body.removeChild(element);
            return void 0;
        }
    });
    // Add documentation link to the help menu
    if (mainmenu)
        mainmenu.helpMenu.addGroup([{ command: documentation }], 2);
}
//# sourceMappingURL=plugin.js.map